﻿using System;
using System.Linq;
using System.Globalization;
using System.IO;
using System.Collections.ObjectModel;
using System.Diagnostics;
using GeoAPI.Geometries;
using NetTopologySuite.Index.Strtree;
using ThinkGeo.MapSuite.Drawing;
using ThinkGeo.MapSuite.Layers;
using ThinkGeo.MapSuite.Shapes;
using ThinkGeo.MapSuite.Styles;

namespace _20190218_TicketViewAnalyseTerre
{
    // IMPORTANT: You have to reference NetTopologySuite.dll & GeoApi.dll in your project.  They come with Map Suite.

    // This class speeds up the loading of a large number of Raster layers by loading and drawing on demand only the files
    // in the current extent.  Normally if you had 100 Raster files you would need to load 100 layers; however, this has performance
    // issues--so we created this high level layer.  It loads a reference file that contains the bounding box, path and file information for 
    // all of the Raster files.  We load this information into an in-memory spatial index. When the map requests to draw the layer, we find the
    // Rasters that are in the current extent, create a layer on-the-fly, call their Draw method and then close them.  In this way, we load
    // on demand only the files that are in the current extent.

    // I have also included a small routine to build a reference file from a directory of Raster files.

    // Reference File Format: [UpperLeftPointX],[LowerRightPoint.X],[UpperLeftPoint.Y],[LowerRightPoint.Y],[Path & File Name to Raster]    

    // Sample:
    // -180.02197265625,-157.52197265625,-67.47802734375,-89.97802734375,..\..\App_Data\RasterImage\1.jpg
    // -112.52197265625,-90.02197265625,-67.47802734375,-89.97802734375,..\..\App_Data\RasterImage\10.jpg
    // 67.47802734375,89.97802734375,67.52197265625,45.02197265625,..\..\App_Data\RasterImage\100.jpg
    // 89.97802734375,112.47802734375,67.52197265625,45.02197265625,..\..\App_Data\RasterImage\101.jpg

    /// <summary>
    /// 
    /// </summary>
    public class MultiGeoRasterLayer : Layer
    {

        private string _rasterReferencePathFileName;
        private STRtree<string> _spatialIndex;
        private RectangleShape _boundingBox;

        private const int UpperLeftXPosition = 0;
        private const int UpperLeftYPosition = 2;
        private const int LowerRightXPosition = 1;
        private const int LowerRightYPosition = 3;
        private const int PathFileNamePosition = 4;


        /// <summary>
        /// 
        /// </summary>
        public MultiGeoRasterLayer()
            : this(string.Empty, double.MaxValue, double.MinValue, string.Empty)
        { }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="rasterReferencePathFileName"></param>
        /// <param name="upperScale"></param>
        /// <param name="lowerScale"></param>
        /// <param name="nomLayer"></param>
        public MultiGeoRasterLayer(string rasterReferencePathFileName, double upperScale, double lowerScale, string nomLayer)
        {
            RasterReferencePathFileName = rasterReferencePathFileName;
            UpperScale = upperScale;
            LowerScale = lowerScale;
            NomLayer = nomLayer;
        }

        /// <summary>
        /// 
        /// </summary>
        public double UpperScale { get; set; }

        /// <summary>
        /// 
        /// </summary>
        public double LowerScale { get; set; }

        public STRtree<string> SpatialIndex
        {
            get { return _spatialIndex; }
            set { _spatialIndex = value; }
        }

        public string RasterReferencePathFileName
        {
            get { return _rasterReferencePathFileName; }
            set { _rasterReferencePathFileName = value; }
        }

        private string nomLayer;
        public string NomLayer
        {
            get { return nomLayer; }
            set { nomLayer = value; }
        }


        // Here on the OpenCore we load our reference file and build the spatial index, which will be used in the DrawCore later.
        // You need to make sure the reference file is in the right format as described in the comments above.
        protected override void OpenCore()
        {
            if (File.Exists(RasterReferencePathFileName))
            {
                var rasterFiles = File.ReadAllLines(RasterReferencePathFileName);

                switch (rasterFiles.Length)
                {
                    case 0:
                        _spatialIndex = new STRtree<string>(rasterFiles.Length + 2);
                        break;
                    case 1:
                        _spatialIndex = new STRtree<string>(rasterFiles.Length + 1);
                        break;

                    default:
                        _spatialIndex = new STRtree<string>(rasterFiles.Length);
                        break;
                }


                Collection<BaseShape> boundingBoxes = new Collection<BaseShape>();

                foreach (string rasterLine in rasterFiles)
                {
                    var parts = rasterLine.Split(new[] { ";" }, StringSplitOptions.None);
                    var rasterBoundingBox = new RectangleShape(new PointShape(double.Parse(parts[UpperLeftXPosition], CultureInfo.InvariantCulture), double.Parse(parts[UpperLeftYPosition], CultureInfo.InvariantCulture)), new PointShape(double.Parse(parts[LowerRightXPosition], CultureInfo.InvariantCulture), double.Parse(parts[LowerRightYPosition], CultureInfo.InvariantCulture)));
                    var envelope = new Envelope(double.Parse(parts[UpperLeftXPosition], CultureInfo.InvariantCulture), double.Parse(parts[LowerRightXPosition], CultureInfo.InvariantCulture), double.Parse(parts[UpperLeftYPosition], CultureInfo.InvariantCulture), double.Parse(parts[LowerRightYPosition], CultureInfo.InvariantCulture));

                    boundingBoxes.Add(rasterBoundingBox);
                    _spatialIndex.Insert(envelope, parts[PathFileNamePosition]);
                }
                _spatialIndex.Build();
                _boundingBox = ExtentHelper.GetBoundingBoxOfItems(boundingBoxes);
            }
            else
            {
                throw new FileNotFoundException("The Raster reference file could not be found.", RasterReferencePathFileName);
            }
        }

        // Here we set the spatial index to null to clean up the memory and get ready for serialization
        protected override void CloseCore()
        {
            _spatialIndex = null;
        }




        /// <summary>
        ///private abstract RasterLayer rasterRasterLayer;
        /// This is just a handy function to build a reference file from a directory.
        ///You can tailor this code to fit your needs.
        /// </summary>
        /// <param name="newReferencepathFileName"></param>
        /// <param name="pathOfRasterFiles"></param>
        /// <param name="inclureSousRépertoires"></param>
        /// <exception cref="DirectoryNotFoundException"></exception>
        public void BuildReferenceFile(string newReferencepathFileName, string pathOfRasterFiles)
        {
            if (Directory.Exists(pathOfRasterFiles))
            {
                var repTxt = Path.GetDirectoryName(newReferencepathFileName);
                if (!Directory.Exists(repTxt))
                    Directory.CreateDirectory(repTxt);

                string[] files;
                string[] extensions = { ".ecw", ".tif", ".jp2" };


                    files = Directory.GetFiles(pathOfRasterFiles, "*.*")
                        .Where(f => extensions.Contains(new FileInfo(f).Extension.ToLower())).ToArray();


                //StreamWriter streamWriter = null;

                using (var streamWriter = File.CreateText(newReferencepathFileName))
                {

                    foreach (var file in files)
                    {

                        var rasterRasterLayer = CreateLayer(file);

                        if (rasterRasterLayer != null)
                        {
                            try
                            {
                                if (NomLayer != null) rasterRasterLayer.Name = NomLayer;
                                if (!rasterRasterLayer.IsOpen) rasterRasterLayer.Open();
                                var boundingBox = rasterRasterLayer.GetBoundingBox();
                                rasterRasterLayer.Close();
                                streamWriter.WriteLine(
                                    $"{boundingBox.UpperLeftPoint.X.ToString(CultureInfo.InvariantCulture)};{boundingBox.LowerRightPoint.X.ToString(CultureInfo.InvariantCulture)};{boundingBox.UpperLeftPoint.Y.ToString(CultureInfo.InvariantCulture)};{boundingBox.LowerRightPoint.Y.ToString(CultureInfo.InvariantCulture)};{file}");

                            }
                            catch (Exception ex)
                            {
                                Trace.TraceError(ex.ToString());
                                Trace.TraceError("Raster défectueux : {0}", file);
                            }
                            finally
                            {

                            }
                        }

                    }

                }
            }
            else
            {
                throw new DirectoryNotFoundException("The path containing the Raster files could not be found.");
            }
        }



        protected override void DrawCore(GeoCanvas canvas, Collection<SimpleCandidate> labelsInAllLayers)
        {
            var currentScale = ExtentHelper.GetScale(canvas.CurrentWorldExtent, canvas.Width, canvas.MapUnit);

            if (currentScale >= LowerScale && currentScale <= UpperScale)
            {
                var currentExtent = canvas.CurrentWorldExtent;
                var currentExtentEnvelope = new Envelope(currentExtent.UpperLeftPoint.X, currentExtent.LowerRightPoint.X, currentExtent.UpperLeftPoint.Y, currentExtent.LowerRightPoint.Y);
                var rasters = SpatialIndex.Query(currentExtentEnvelope);

                //System.Diagnostics.Debug.Print("MultiGeoRasterLayer.DrawCore: extent={0}, matches={1}", currentExtent.ToString(), rasters.Count);

                foreach (var file in rasters)
                {

                    var rasterRasterLayer = CreateLayer(file);

                    try
                    {
                        if (!rasterRasterLayer.IsOpen) rasterRasterLayer.Open();
                        rasterRasterLayer.Draw(canvas, labelsInAllLayers);
                        rasterRasterLayer.Close();
                    }
                    catch (Exception e)
                    {
                        System.Diagnostics.Trace.TraceError(e.ToString());
                    }
                }
            }
        }

        public override bool HasBoundingBox
        {
            get => true;
        }

        //  We use the cached bounding box we set in the OpenCore  
        protected override RectangleShape GetBoundingBoxCore()
        {
            return _boundingBox;
        }

        private static RasterLayer CreateLayer(string file)
        {
            var extension = Path.GetExtension(file);

            switch (extension.ToLower())
            {
                //case ".ecw": return new EcwRasterLayer(file);
                //case ".tif": return new GeoTiffRasterLayer(file);
                //case ".jp2": return new Jpeg2000RasterLayer(file);
                default: return null;
            }
        }

        //private static NativeImageRasterLayer CreateLayerBis(string file)
        //{
        //    var extension = Path.GetExtension(file);

        //    switch (extension.ToLower())
        //    {
        //        case ".ecw": return new NativeImageRasterLayer(file);
        //        case ".tif": return new NativeImageRasterLayer(file);
        //        case ".jp2": return new NativeImageRasterLayer(file);
        //        default: return null;
        //    }
        //}


    }
}
